#--
# Copyright (c) 2009-2010 Novell, Inc.
#
# All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of version 2 of the GNU General Public License
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, contact Novell, Inc.
#
# To contact Novell about this file by physical or electronic mail,
# you may find current contact information at www.novell.com
#++

require "install_in_progress_exception"
require "license_required_exception"

class PatchesController < ApplicationController
  include ERB::Util

  # just for marking the kind texts for translation, used in _patches.html.erb
  # TRANSLATORS: patch severity
  KIND_TRANSLATIONS = [ N_("Low"), N_("Enhancement"), N_("Normal"),
    N_("Bugfix"), N_("Recommended"), N_("Security"), N_("Other") ]

private

  def collect_done_patches
    return Rails.cache.fetch("patch:installed") || []
  end

  def read_messages
    if File.exists?(Patch::MESSAGES_FILE)
      msg = File.read(Patch::MESSAGES_FILE)
      return [{:message => msg}]
    end

    return []
  end

  def refresh_timeout
    # the default is 1 hours
    timeout = ControlPanelConfig.read 'patch_status_timeout', 1.hour

    if timeout.zero?
      Rails.logger.info "Patch status autorefresh is disabled"
    else
      Rails.logger.info "Autorefresh patch status after #{timeout} seconds"
    end

    return timeout
  end

  # remove the last cached result
  def reset_results
    Rails.logger.info "Clearing patch result cache"
    Rails.cache.delete("patch:installed")
    Rails.cache.delete("patch:failed")
  end

  public

  # GET /patches
  # GET /patches.xml
  def index
    authorize! :read, Patch

    @msgs = read_messages
    if params['messages']
      Rails.logger.debug "Reading patch messages"
      respond_to do |format|
        format.xml { render  :xml => @msgs.to_xml( :root => "messages", :dasherize => false ) }
        format.json { render :json => @msgs.to_json( :root => "messages", :dasherize => false ) }
      end
      return
    end

    if !@msgs.blank? && request.format.html?
      #@msgs.gsub!('<br/>', ' ')
      @message = @msgs.map{|s| s[:message]}.to_s.gsub(/\n/, '')
      # FIXME: do NOT put messages to flash, the storage is VERY limited (< 4kB), just put a link to show them
      flash[:warning] = _("There are patch installation messages available") + details(@message)
    end

    # check if a license is required
    if PatchesState.read[:message_id] == "PATCH_EULA"
      respond_to do |format|
         #move to page for license confirmation
        format.html { redirect_to :action => "license" }
        format.xml  { render :xml => LicenseRequiredException.new.to_xml }
        format.json { render :json => LicenseRequiredException.new.to_json }
      end

      return
    end

    #checking if an installation is running
    running, remaining = Patch.installing
    if running #there is process which runs installation
      unless request.format.html?
        respond_to do |format|
          format.xml  { render :xml => InstallInProgressException.new(remaining).to_xml }
          format.json { render :json => InstallInProgressException.new(remaining).to_json }
        end

        return
      end

      # patch update installation in progress
      # display the message and reload after a while
      @flash_message = h _("Cannot read available patches, patch installation is in progress.")

      if remaining.present?
        @flash_message << "<br/>".html_safe
        @flash_message << h(n_("There is one patch to install.", "There are %d patches to install.", remaining) % remaining)
      end

      @patch_updates = []
      @error = true
      @reload = true
    else
      #no installation process
      begin
        @patch_updates = Patch.find(:all)
      rescue Exception => e
        Rails.logger.warn "Error while reading patches: #{e.inspect}"
        if e.message.match /Repository (.*) needs to be signed/
          flash[:error] = ((h _("Cannot read patch updates: GPG key for repository %s is not trusted.")) % "<em>#{h $1}</em>").html_safe
        elsif e.message.match /System management is locked by the application with pid ([0-9]+) \((.*)\)\./
          flash[:warning] = _("Software management is locked by another application ('%s', PID %s).") % [$2, $1]
        elsif e.message.match /The ZYpp package manager is locked by process ([0-9]+)\. Retry later\./
          flash[:warning] = _("Software management is locked by another application (PID %s). Available patches cannot be read.") % $1
        end
        @patch_updates = []
        @error = true
      end
      @patch_updates = @patch_updates + collect_done_patches #report also which patches is installed
    end
    logger.info "All patches: #{@patch_updates.inspect}"
    respond_to do |format|
      format.html {}
      format.xml { render  :xml => @patch_updates.to_xml( :root => "patches", :dasherize => false ) }
      format.json { render :json => @patch_updates.to_json( :root => "patches", :dasherize => false ) }
    end
  end

  # GET /patches/1
  # GET /patches/1.xml
  def show
    authorize! :read, Patch
    # RORSCAN_INL: User has already read permission for ALL patch info
    @patch_update = Patch.find(params[:id])
    if @patch_update.blank?
      logger.error "Patch #{params[:id]} not found."
      render ErrorResult.error(404, 1, "Patch #{params[:id]} not found.") and return
    end
    respond_to do |format|
      format.xml { render  :xml => @patch_update.to_xml(:dasherize => false, :indent => 2) }
      format.json { render :json => @patch_update.to_json(:dasherize => false, :indent => 2) }
    end
  end

  # this action is rendered as a partial, so it can't throw
  def show_summary
    if cannot? :read, Patch
      render :text => _("Status not available (no permissions)") and return
    end

    error = nil
    patch_updates = nil
    ref_timeout = nil
    error_type = :none

    running, remaining = Patch.installing

    if running
      ref_timeout = 10.seconds
      error_type = :install
      error_string = h _("Patch installation is in progress.")

      if remaining.present?
        error_string << "<br/>".html_safe
        error_string << h(n_("There is one patch to install.", "There are %d patches to install.", remaining) % remaining)
      end
    elsif PatchesState.read[:message_id] == "PATCH_EULA" #checking if there is a missing licence
      error_type = :license
    else
      #evaluate available patches
      begin
        patch_updates = Patch.find :all
        patch_updates = patch_updates + collect_done_patches #report also which patches is installed
        ref_timeout = refresh_timeout
      rescue Exception => error
        Rails.logger.warn "Caught exception: #{error.inspect}"

        if error.message.match /Repository (.*) needs to be signed/
          error_string = (_("Cannot read patch updates: GPG key for repository <em>%s</em> is not trusted.") % h($1)).html_safe
          error_type = :unsigned
        elsif error.message.match /System management is locked by the application with pid ([0-9]+) \((.*)\)\./
          error_string = _("Software management is locked by another application ('%s', PID %s).") % [$2, $1]
          error_type = :locked
        elsif error.message.match /The ZYpp package manager is locked by process ([0-9]+)\. Retry later\./
          error_string = _("Software management is locked by another application (PID %s).") % $1
          error_type = :locked
        else
          error_string = error.message
          error_type = :unknown
        end
      end
    end

    patches_summary = nil
    Rails.logger.debug "patch_updates: #{patch_updates.inspect}"
    Rails.logger.debug "installed patches: #{Rails.cache.fetch("patch:installed").inspect}"
    Rails.logger.debug "failed patches: #{Rails.cache.fetch("patch:failed").inspect}"

    if patch_updates
      patches_summary = { :security => 0, :important => 0, :optional => 0}
      [:security, :important, :optional].each do |patch_type|
        patches_summary[patch_type] = patch_updates.find_all { |p| p.kind == patch_type.to_s && !p.installed }.size
      end
      # add 'low' patches to optional
      patches_summary[:optional] += patch_updates.find_all { |p| p.kind == 'low' && !p.installed }.size
      # add 'normal' patches to important
      patches_summary[:important] += patch_updates.find_all { |p| p.kind == 'normal' && !p.installed }.size
    else
      flash.clear #no flash from load_proxy
    end

    Rails.logger.debug "Patch summary: #{{:patch => patches_summary, :error => error,
     :error_string => error_string, :error_type => error_type, :refresh_timeout => ref_timeout}.inspect}"

    respond_to do |format|
      format.html { render :partial => "patch_summary",
                           :locals => { :patch => patches_summary,
                           :error => error,
                           :error_string => error_string,
                           :error_type => error_type,
                           :refresh_timeout => ref_timeout } }
      # TODO error handling in JSON and XML
      format.json  { render :json => patches_summary }
      format.xml  { render :xml => patches_summary }
    end
  end

  def load_filtered
    authorize! :read, Patch
    @patch_updates = Patch.find :all
    kind = params[:value]
    search_map = { "green" => ["normal","low","bugfix","recommended"], "security" => ["security"],
                   "important" => "important", 'optional' => ["enhancement","other"] }
    unless kind == "all"
      search_list = search_map[kind] || [kind]
      @patch_updates = @patch_updates.find_all { |patch| search_list.include?(patch.kind) }
    end
    render :partial => "patches"
  end


  # POST /patches/start_install_all
  # Starting installation of all proposed patches
  def start_install_all
    authorize! :install, Patch

    Patch::BM.background_enabled? ? Patch.install_all_background : Patch.install_all

    show_summary
  end

  # POST /patches/install
  # Installing one or more patches which has given via param

  def install
    authorize! :install, Patch

    running, remaining = Patch.installing

    if running
      if request.format.html?
        error_string = _("Patch installation is in progress.")

        if remaining.present?
          error_string << " "
          error_string << n_("There is one patch to install.", "There are %d patches to install.", remaining) % remaining
        end

        flash[:error] = error_string
        render :show
      else
        respond_to do |format|
          format.xml  { render :xml => InstallInProgressException.new(remaining).to_xml }
          format.json { render :json => InstallInProgressException.new(remaining).to_json }
        end
      end

      return
    end

    update_array = []

    #search for patches and collect the ids
    params.each { |key, value|
      if key.start_with?("patch_") || key == "id"
        update_array << value
      end
    }
    @patch_update = Patch.new({})

    # remove the last cached results
    reset_results

    begin
      if !update_array.empty?
        Patch::BM.background_enabled? ? Patch.install_patches_by_id_background(update_array) : Patch.install_patches_by_id(update_array)
      end
    # FIXME: this might hide some errors
    rescue Exception => e
      Rails.logger.info "Some patches are not needed in #{update_array.inspect} anymore: #{e.message}"
    end

    respond_to do |format|
      format.html {
        logger.debug "*** Check before redirect: basesystem setup completed -> #{Basesystem.new.load_from_session(session).completed?}"

        if Basesystem.new.load_from_session(session).completed?
          redirect_to :controller => "controlpanel", :action => "index" and return
        else
          redirect_to :controller => "controlpanel", :action => "nextstep" and return
        end
      }

      # return 202 Accepted (processing in background) or 200 OK (direct processing)
      format.json { head Patch::BM.background_enabled? ? :accepted : :success }
      format.xml  { head Patch::BM.background_enabled? ? :accepted : :success }
    end
  end

  def message
    authorize! :read, Patch
    authorize! :install, Patch

    logger.warn "Confirmation of reading patch messages"
    File.delete Patch::MESSAGES_FILE

    respond_to do |format|
      format.html {
        redirect_to "/"
      }
      format.xml  { head :ok }
      format.json { head :ok }
    end
  end

  def update_license
    authorize! :read, Patch
    authorize! :install, Patch

    if (params[:accept].present? || params[:reject].present?) && params[:name].present? && params[:patch_id].present?
      params[:accept].present? ? Patch.accept_license(params[:name]) : Patch.reject_license(params[:name])

      if params[:accept].present?
        Rails.logger.info "Accepted license for patch: #{params[:patch_id]}"
        Patch.accept_eulas

        # remove the previous patch installation failure caused by not accepted license
        failed_patches = Rails.cache.fetch("patch:failed")
        unless failed_patches.blank?
          failed_patches.delete_if {|patch| patch.resolvable_id == params[:patch_id]}
          Rails.cache.write("patch:failed", failed_patches)
        end

        patches = [params[:patch_id]]
        skipped_patches = Rails.cache.fetch("patch:skipped") {[]}
        unless skipped_patches.empty?
          Rails.logger.info "Found #{skipped_patches.size} skipped patches"
          patches += skipped_patches
        end

        # install the patches again
        Patch::BM.background_enabled? ? Patch.install_patches_by_id_background(patches) : Patch.install_patches_by_id(patches)
      end

      if request.format.html?
        redirect_to "/"
      else
        @patch_update = Patch.new({})
        render :show
      end
      return
    else
      if request.format.html?
        flash[:error] = _("The license has not been accepted nor rejected.")
        redirect_to root_path
      else
        # Error 422
        head :unprocessable_entity
      end
    end
  end

  def license
    authorize! :read, Patch
    authorize! :install, Patch

    respond_to do |format|
      format.html {
        @license = Patch.license.first
        @text = @license[:text] if @license
        @package_id = @license[:package_id] if @license
        @patch_id = @license[:patch_id] if @license
        if @text =~ /DT:Rich/ #text is richtext for packager
          @text = @text.html_safe
        else
          (h @text).gsub(/^\n\n$/, '<p class="eula_paragraph">').html_safe
        end
        render
      }
      format.xml { render  :xml => Patch.license.to_xml( :root => "licenses", :dasherize => false ) }
      format.json { render :json => Patch.license.to_json( :root => "licenses", :dasherize => false ) }
    end
  end


end
